@@ -1,24 +1,19 @@
-from __future__ import annotations
-
-import dataclasses
-import traceback
-import warnings
-from types import TracebackType
-
-from ordered_set import OrderedSet
-
-from .util import get_qualified_name
-
 """
 Flask-Limiter Extension
 """
+from __future__ import annotations
+
+import dataclasses
 import datetime
 import itertools
 import logging
 import time
+import traceback
+import warnings
 import weakref
 from collections import defaultdict
 from functools import partial, wraps
+from types import TracebackType
 from typing import Type, overload
 
 import flask
@@ -26,6 +21,7 @@ import flask.wrappers
 from limits.errors import ConfigurationError
 from limits.storage import MemoryStorage, Storage, storage_from_string
 from limits.strategies import STRATEGIES, RateLimiter
+from ordered_set import OrderedSet
 from werkzeug.http import http_date, parse_date
 
 from ._compat import request_context
@@ -46,12 +42,12 @@ from .typing import (
     Union,
     cast,
 )
+from .util import get_qualified_name
 from .wrappers import Limit, LimitGroup, RequestLimit
 
 
 @dataclasses.dataclass
 class LimiterContext:
-    rate_limiting_complete: dict[str, bool] = dataclasses.field(default_factory=dict)
     view_rate_limit: Optional[RequestLimit] = None
     view_rate_limits: List[RequestLimit] = dataclasses.field(default_factory=list)
     conditional_deductions: Dict[Limit, List[str]] = dataclasses.field(
@@ -60,10 +56,10 @@ class LimiterContext:
     seen_limits: OrderedSet[Limit] = dataclasses.field(default_factory=OrderedSet)
 
     def reset(self) -> None:
-        self.rate_limiting_complete.clear()
         self.view_rate_limit = None
         self.view_rate_limits.clear()
         self.conditional_deductions.clear()
+        self.seen_limits.clear()
 
 
 class Limiter:
@@ -74,8 +70,8 @@ class Limiter:
       by.
     :param app: :class:`flask.Flask` instance to initialize the extension with.
     :param default_limits: a variable list of strings or callables
-     returning strings denoting default limits to apply to all routes.
-     :ref:`ratelimit-string` for  more details.
+     returning strings denoting default limits to apply to all routes that are
+     not explicitely decorated with a limit. :ref:`ratelimit-string` for  more details.
     :param default_limits_per_method: whether default limits are applied
      per method, per route or as a combination of all method per route.
     :param default_limits_exempt_when: a function that should return
@@ -89,6 +85,13 @@ class Limiter:
     :param application_limits: a variable list of strings or callables
      returning strings for limits that are applied to the entire application
      (i.e a shared limit for all routes)
+    :param application_limits_per_method: whether application limits are applied
+     per method, per route or as a combination of all method per route.
+    :param application_limits_exempt_when: a function that should return
+     True/False to decide if the application limits should be skipped
+    :param application_limits_deduct_when: a function that receives the
+     current :class:`flask.Response` object and returns True/False to decide
+     if a deduction should be made from the application rate limit(s)
     :param application_limits_cost: The cost of a hit to the global application
      limits as an integer or a function that takes no parameters and returns an
      integer (Default: ``1``).
@@ -121,6 +124,8 @@ class Limiter:
     :param retry_after: Allows configuration of how the value of the
      `Retry-After` header is rendered. One of `http-date` or `delta-seconds`.
     :param key_prefix: prefix prepended to rate limiter keys and app context global names.
+    :param request_identifier: a callable that returns the unique identity the current request.
+     Defaults to :attr:`flask.Request.endpoint`
     :param enabled: Whether the extension is enabled or not
     """
 
@@ -137,6 +142,11 @@ class Limiter:
         ] = None,
         default_limits_cost: Optional[Union[int, Callable[[], int]]] = None,
         application_limits: Optional[List[Union[str, Callable[[], str]]]] = None,
+        application_limits_per_method: Optional[bool] = None,
+        application_limits_exempt_when: Optional[Callable[[], bool]] = None,
+        application_limits_deduct_when: Optional[
+            Callable[[flask.wrappers.Response], bool]
+        ] = None,
         application_limits_cost: Optional[Union[int, Callable[[], int]]] = None,
         headers_enabled: Optional[bool] = None,
         header_name_mapping: Optional[Dict[HeaderNames, str]] = None,
@@ -153,6 +163,7 @@ class Limiter:
         in_memory_fallback_enabled: Optional[bool] = None,
         retry_after: Optional[str] = None,
         key_prefix: str = "",
+        request_identifier: Optional[Callable[..., str]] = None,
         enabled: bool = True,
     ) -> None:
         self.app = app
@@ -164,6 +175,9 @@ class Limiter:
         self._default_limits_exempt_when = default_limits_exempt_when
         self._default_limits_deduct_when = default_limits_deduct_when
         self._default_limits_cost = default_limits_cost
+        self._application_limits_per_method = application_limits_per_method
+        self._application_limits_exempt_when = application_limits_exempt_when
+        self._application_limits_deduct_when = application_limits_deduct_when
         self._application_limits_cost = application_limits_cost
         self._in_memory_fallback = []
         self._in_memory_fallback_enabled = in_memory_fallback_enabled or (
@@ -190,6 +204,7 @@ class Limiter:
 
         self._key_func = key_func
         self._key_prefix = key_prefix
+        self._request_identifier = request_identifier
 
         _default_limits = (
             [
@@ -338,11 +353,25 @@ class Limiter:
         )
 
         self._key_prefix = self._key_prefix or config.get(ConfigVars.KEY_PREFIX, "")
-
+        self._request_identifier = self._request_identifier or config.get(
+            ConfigVars.REQUEST_IDENTIFIER, lambda: flask.request.endpoint or ""
+        )
         app_limits = config.get(ConfigVars.APPLICATION_LIMITS, None)
         self._application_limits_cost = self._application_limits_cost or config.get(
             ConfigVars.APPLICATION_LIMITS_COST, 1
         )
+        if self._application_limits_per_method is None:
+            self._application_limits_per_method = bool(
+                config.get(ConfigVars.APPLICATION_LIMITS_PER_METHOD, False)
+            )
+        self._application_limits_exempt_when = (
+            self._application_limits_exempt_when
+            or config.get(ConfigVars.APPLICATION_LIMITS_EXEMPT_WHEN)
+        )
+        self._application_limits_deduct_when = (
+            self._application_limits_deduct_when
+            or config.get(ConfigVars.APPLICATION_LIMITS_DEDUCT_WHEN)
+        )
 
         if not self.limit_manager._application_limits and app_limits:
             self.limit_manager.set_application_limits(
@@ -352,6 +381,9 @@ class Limiter:
                         key_function=self._key_func,
                         scope="global",
                         shared=True,
+                        per_method=self._application_limits_per_method,
+                        exempt_when=self._application_limits_exempt_when,
+                        deduct_when=self._application_limits_deduct_when,
                         cost=self._application_limits_cost,
                     )
                 ]
@@ -360,6 +392,9 @@ class Limiter:
             app_limits = self.limit_manager._application_limits
             for group in app_limits:
                 group.cost = self._application_limits_cost
+                group.per_method = self._application_limits_per_method
+                group.exempt_when = self._application_limits_exempt_when
+                group.deduct_when = self._application_limits_deduct_when
             self.limit_manager.set_application_limits(app_limits)
 
         conf_limits = config.get(ConfigVars.DEFAULT_LIMITS, None)
@@ -398,13 +433,22 @@ class Limiter:
 
     @property
     def context(self) -> LimiterContext:
+        """
+        The context is meant to exist for the lifetime
+        of a request/response cycle per instance of the extension
+        so as to keep track of any state used at different steps
+        in the lifecycle (for example to pass information
+        from the before request hook to the after_request hook)
+
+        :meta private:
+        """
         ctx = request_context()
         if not hasattr(ctx, "_limiter_request_context"):
             ctx._limiter_request_context = defaultdict(LimiterContext)  # type: ignore
         return cast(
-            Dict[str, LimiterContext],
+            Dict[Limiter, LimiterContext],
             ctx._limiter_request_context,  # type: ignore
-        )[self._key_prefix]
+        )[self]
 
     def limit(
         self,
@@ -777,6 +821,19 @@ class Limiter:
 
         return self.context.view_rate_limits
 
+    def identify_request(self) -> str:
+        """
+        Returns the identity of the request (by default this is the
+        :attr:`flask.Request.endpoint` associated by the view function
+        that is handling the request). The behavior can be customized
+        by initializing the extension with a callable argument for
+        :paramref:`~flask_limiter.Limiter.request_identifier`.
+        """
+        if self.initialized and self.enabled:
+            assert self._request_identifier
+            return self._request_identifier()
+        return ""
+
     def __check_conditional_deductions(self, response: flask.wrappers.Response) -> None:
         for lim, args in self.context.conditional_deductions.items():
             if lim.deduct_when and lim.deduct_when(response):
@@ -785,7 +842,7 @@ class Limiter:
                 except Exception as err:
                     if self._swallow_errors:
                         self.logger.exception(
-                            "Failed to deduct rate limit. " "Swallowing error"
+                            "Failed to deduct rate limit. Swallowing error"
                         )
                     else:
                         raise err
@@ -808,9 +865,12 @@ class Limiter:
                     str(header_limit.limit.amount),
                 )
                 response.headers.add(
-                    self._header_mapping[HeaderNames.REMAINING], header_limit.remaining
+                    self._header_mapping[HeaderNames.REMAINING],
+                    str(header_limit.remaining),
+                )
+                response.headers.add(
+                    self._header_mapping[HeaderNames.RESET], str(reset_at)
                 )
-                response.headers.add(self._header_mapping[HeaderNames.RESET], reset_at)
 
                 # response may have an existing retry after
                 existing_retry_after_header = response.headers.get("Retry-After")
@@ -834,9 +894,11 @@ class Limiter:
                 # set the header instead of using add
                 response.headers.set(
                     self._header_mapping[HeaderNames.RETRY_AFTER],
-                    self._retry_after == "http-date"
-                    and http_date(reset_at)
-                    or int(reset_at - time.time()),
+                    str(
+                        http_date(reset_at)
+                        if self._retry_after == "http-date"
+                        else int(reset_at - time.time())
+                    ),
                 )
             except Exception as e:  # noqa: E722
                 if self._in_memory_fallback_enabled and not self._storage_dead:
@@ -857,18 +919,14 @@ class Limiter:
         return response
 
     def __check_all_limits_exempt(
-        self, endpoint: Optional[str], callable_name: Optional[str] = None
+        self,
+        endpoint: Optional[str],
     ) -> bool:
         return bool(
             not endpoint
             or not (self.enabled and self.initialized)
-            or endpoint == "static"
+            or endpoint.split(".")[-1] == "static"
             or any(fn() for fn in self._request_filters)
-            or (
-                self.context.rate_limiting_complete.get(callable_name, False)
-                if callable_name
-                else any(self.context.rate_limiting_complete.values())
-            )
         )
 
     def __filter_limits(
@@ -878,14 +936,13 @@ class Limiter:
         callable_name: Optional[str],
         in_middleware: bool = False,
     ) -> List[Limit]:
-
         if callable_name:
             name = callable_name
         else:
             view_func = flask.current_app.view_functions.get(endpoint or "", None)
             name = get_qualified_name(view_func) if view_func else ""
 
-        if self.__check_all_limits_exempt(endpoint, callable_name):
+        if self.__check_all_limits_exempt(endpoint):
             return []
 
         marked_for_limiting = (
@@ -908,18 +965,20 @@ class Limiter:
                     self.__check_backend_count = 0
                 else:
                     fallback_limits = list(itertools.chain(*self._in_memory_fallback))
-        resolved_limits = self.limit_manager.resolve_limits(
+        if fallback_limits:
+            return fallback_limits
+
+        defaults, decorated = self.limit_manager.resolve_limits(
             flask.current_app,
             endpoint,
             blueprint,
             name,
             in_middleware,
             marked_for_limiting,
-            fallback_limits,
         )
-        limits = OrderedSet(resolved_limits) - self.context.seen_limits
-        self.context.seen_limits.update(limits)
-        return list(limits)
+        limits = OrderedSet(defaults) - self.context.seen_limits
+        self.context.seen_limits.update(defaults)
+        return list(limits) + list(decorated)
 
     def __evaluate_limits(self, endpoint: str, limits: List[Limit]) -> None:
         failed_limits: List[Tuple[Limit, List[str]]] = []
@@ -951,10 +1010,8 @@ class Limiter:
                 method = self.limiter.hit
                 kwargs["cost"] = lim.cost
 
-            if not limit_for_header or lim.limit < limit_for_header.limit:
-                limit_for_header = RequestLimit(self, lim.limit, args, False)
-
-            view_limits.append(RequestLimit(self, lim.limit, args, False))
+            request_limit = RequestLimit(self, lim.limit, args, False, lim.shared)
+            view_limits.append(request_limit)
 
             if not method(lim.limit, *args, **kwargs):
                 self.logger.info(
@@ -969,12 +1026,21 @@ class Limiter:
                 if self._fail_on_first_breach:
                     break
 
+        if not limit_for_header and view_limits:
+            # Pick a non shared limit over a shared one if possible
+            # when no rate limit has been hit. This should be the best hint
+            # for the client.
+            explicit = [limit for limit in view_limits if not limit.shared]
+            limit_for_header = explicit[0] if explicit else view_limits[0]
+
         self.context.view_rate_limit = limit_for_header or None
         self.context.view_rate_limits = view_limits
 
         on_breach_response = None
         for limit in failed_limits:
-            request_limit = RequestLimit(self, limit[0].limit, limit[1], True)
+            request_limit = RequestLimit(
+                self, limit[0].limit, limit[1], True, limit[0].shared
+            )
             for cb in dict.fromkeys([self._on_breach, limit[0].on_breach]):
                 if cb:
                     try:
@@ -997,10 +1063,10 @@ class Limiter:
     def _check_request_limit(
         self, callable_name: Optional[str] = None, in_middleware: bool = True
     ) -> None:
-        endpoint = flask.request.endpoint or ""
+        endpoint = self.identify_request()
         try:
             all_limits = self.__filter_limits(
-                flask.request.endpoint,
+                endpoint,
                 flask.request.blueprint,
                 callable_name,
                 in_middleware,
@@ -1027,8 +1093,7 @@ class Limiter:
                     raise e
 
     def __release_context(self, _: Optional[BaseException] = None) -> None:
-        if self.context:
-            self.context.reset()
+        self.context.reset()
 
 
 class LimitDecorator:
@@ -1100,7 +1165,7 @@ class LimitDecorator:
         )
 
         self.limiter.limit_manager.add_endpoint_hint(
-            flask.request.endpoint, qualified_location
+            self.limiter.identify_request(), qualified_location
         )
 
         self.limiter._check_request_limit(
@@ -1142,24 +1207,26 @@ class LimitDecorator:
             def __inner(*a: P.args, **k: P.kwargs) -> R:
                 if (
                     self.limiter._auto_check
-                    and not self.limiter.context.rate_limiting_complete.get(name, False)
+                    and not getattr(obj, "__wrapper-limiter-instance", None)
+                    == self.limiter
                 ):
-                    if flask.request.endpoint:
-                        view_func = flask.current_app.view_functions.get(
-                            flask.request.endpoint, None
-                        )
+                    identity = self.limiter.identify_request()
+                    if identity:
+                        view_func = flask.current_app.view_functions.get(identity, None)
                         if view_func and not get_qualified_name(view_func) == name:
-                            self.limiter.limit_manager.add_endpoint_hint(
-                                flask.request.endpoint, name
-                            )
+                            self.limiter.limit_manager.add_endpoint_hint(identity, name)
 
                     self.limiter._check_request_limit(
                         in_middleware=False, callable_name=name
                     )
 
-                    self.limiter.context.rate_limiting_complete[name] = True
-                return cast(
-                    R, flask.current_app.ensure_sync(cast(Callable[P, R], obj))(*a, **k)
-                )
+                return cast(R, flask.current_app.ensure_sync(obj)(*a, **k))
 
+            # mark this wrapper as wrapped by a decorator from the limiter
+            # from which the decorator was created. This ensures that stacked
+            # decorations only trigger rate limiting from the inner most
+            # decorator from each limiter instance (the weird need for
+            # keeping track of the instance is to handle cases where multiple
+            # limiter extensions are registered on the same application).
+            setattr(__inner, "__wrapper-limiter-instance", self.limiter)
             return __inner
